!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! TAMES  Topological Analysis of Metabolic Structures                 !
! Copyright (C) 2017                                                   !
! Paulo A. N. Dias, paulo.dias29@gmail.com                             !
! Marco Seabra dos Reis, marco@eq.uc.pt                                !
! Armindo Salvador, armindo.salvador@gmail.com                         !
! Pedro Martins, pmartins@iscac.pt                                     !
!                                                                      !
! This file is part of TAMES.                                          !
!                                                                      !
! TAMES is free software: you can redistribute it and/or modify        !
! it under the terms of the GNU General Public License as published by !
! the Free Software Foundation, either version 3 of the License, or    !
! any later version.                                                   !
!                                                                      !
! You should have received a copy of the GNU General Public License    !
! along with TAMES.  If not, see <http://www.gnu.org/licenses/>.       !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!                                                         !
! Program "StAl-InstCount":                               !
! - Counting the stand-alone instances in the metabolic   !
!   network of the structures (no overlaps or overlap of  !
!   a maximum of 1 metabolite with instances of the same  !
!   structure) within a given topological scope.          !
!                                                         !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

      Program StAlInstCount

! CONSTANTS
!      n1 : number of metabolites
!      n2 : number of reactions (reversibility not considered)
!      repeat : logical variable used to determine the number of reactions and of metabolites in the files
!      maxNumGen : maximum number of generations considered to generate structures
!      maxNumMet : maximum number of metabolites considered for the structures within the considered number of
!                  generations maxNumGen
!      maxNumReact : maximum number of reactions considered for the structures within the considered number of
!                    generations maxNumGen
!      numStructures : number of structures for the considered generations (maxNumGen)
!      numInstReal :  number of instances for the considered scope
!
!
! VECTORS
!      mn(i) : list of names of all metabolites in file "List_Metabolites.txt"
!      rn(i) : list of names of all reactions in file "List_Reactions.txt"
!      sup(i1) : support vector of size integer*2
!      nrefStruLow(i1) : lower reference number of a structure of the i1 generation
!      nrefStruHigh(i1) : higher reference number of a structure of the i1 generation
!      nrefInstLow(i1) : lower reference number of a instance of structure i1 in organized matrices,
!                        OrgMetInst(i2,i1) and OrgReactInst(i2,i1), by structure to store the instances
!      nrefInstHigh(i1) : lower reference number of a instance of structure i1 in organized matrices
!                         OrgMetInst(i2,i1) and OrgReactInst(i2,i1), by structure to store the instances
!      StAlInst(i1) : Indicates if the instance shares metabolites with other instances of the same structure.
!                     The cell can be filled with: 0 => does not share reactions or metabolite with other instances;
!                     1 => shares 1 metabolite with at least one other instance; 2 => shares reactions or more than
!                     one metabolite with other instances
!      const(i1) : support vector with the dimension integer*4
!      EvaInstStru(i1) : number of instances of structure already evaluated for overlaps. Used to help
!                        accelerate the algorithm
!      supResultsRevStruct(i1) : support vector with the number of instances of the structures. Used to obtain
!                                a frequency ranking for the structures
!      supRankStru(i1) : rank of the number of instances of the structures
!
!
! MATRICES
!      DataRevStru(i2,i1) : i1) reference number of the structure. i2) addtional data for each structure: (i2=1) number
!                           of generation; (i2=2) number of metabolites; (i2=3) number of reactions; (i2=4) number of
!                           reversible reactions; (i2=5) number of Ancestors; (i2=6) number of Parents; (i2=7) number of
!                           Childs; (i2=8) number of Offspring; (i2=9) number of interconnections; (i2=10) maximum possible
!                           number of interconnections; (i2=11) number of connections metabolite/reaction (reversibility not
!                           considered); (i2=12) maximum possible number of connections metabolite/reaction (reversibility
!                           not considered)
!      EquiMetRevStru(i2,i1) : information regarding the equivalent metabolites of a given structure i1. If two metabolites are
!                              equivalent, the corresponding cell is filled with "1"; if not, is filled with "0". The first
!                              metabolite is compared with the ones in higher positions, the second metabolite is compared to
!                              metabolites who are also in higher positions, ... till we compare the last two metabolites.
!                              For the metabolite in a general position X (lower than the last position available), it will
!                              be compared to metabolites in increasing higher positions in the cell. Considering the case
!                              where the metabolite in the position X is compared to the metabolite in the position Y, the
!                              information regarding the topological equivalence of the two nodes will be stored in the cell:
!                              i2 = (Somatrio[i=1,(X-1)] n Met. - i ) + ( Y - X )
!      EquiReactRevStru(i2,i1) : information regarding the equivalent reactions of a given structure i1. If two reactions are
!                                equivalent, the corresponding cell is filled with "1"; if not, is filled with "0". The first
!                                reaction is compared with the ones in higher positions, the second reaction is compared to
!                                reactions who are also in higher positions, ... till we compare the last two reactions.
!                                For the reaction in a general position X (lower than the last position available), it will
!                                be compared to reactions in increasing higher positions in the cell. Considering the case
!                                where the reaction in the position X is compared to the reaction in the position Y, the
!                                information regarding the topological equivalence of the two nodes will be stored in the cell:
!                                i2 = (Somatrio[i=1,(X-1)] n Met. - i ) + ( Y - X )
!      SupEquiMetStru(i2,i1) : support matrix that gives the information if metabolite i2 is topologically equivalent to
!                              another metabolite in a lower position of structure i1
!      SupEquiReactStru(i2,i1) : support matrix that gives the information if reaction i2 is topologically equivalent to
!                                another reaction in a lower position of structure i1
!      OrgMetInst(i2,i1) :  reference numbers of the metabolites of the instance i1. The instances are organized according
!                           to their structure: instances of structures of lower reference numbers are stored first
!      OrgReactInst(i2,i1) :  reference numbers of the reactions of the instance i1. The instances are organized according
!                             to their structure: instances of structures of lower reference numbers are stored first
!      InclRevStru(i2,i1) : inclusion matrix. If the structure i1 can be found within structure i2, the corresponding cell
!                           is filled with "1"; if not, the cell is filled with "0"
!      ResultsRevStruct(i2,i1) : output data obtained for the structure i1 with the developed code. (i2=1) "naked" frequency (number of
!                                instances); (i2=2) "embedded" frequency; (i2=3) n of stand-alone instances of type i - 0 metabolites
!                                and 0 reactions shared with other instances of the same structure; (i2=4) n of stand-alone instances
!                                of type ii - 1 metabolite and 0 reactions shared with other instances of the same structure; (i2=5) total
!                                n stand-alone instances - type i + ii; (i2=6) n of different metabolites; (i2=7) n of different reactions
!      MetRevStruct(i2,i1) : frequency of the metabolite i2 in the instances of structure i1
!      ReactRevStruct(i2,i1) : frequency of the reaction i2 in the instances of structure i1
!      MetABCRevStruct(i2,i1) : frequency of metabolite i2 in the instances of a given structure at all the
!                               possible positions. For a given structure X, the frequency of the metabolite
!                               i2 in the position Y is: i1 = [(X - 1) x maxNumMet] + Y, where Y => [1,maxNumMet]
!      React123RevStruct(i2,i1) : frequency of reaction i2 in the instances of a given structure at all the
!                                 possible positions. For a given structure X, the frequency of the reaction
!                                 i2 in the position Y is: i1 = [(X - 1) x (maxNumReact)] + Y, where Y => [1,maxNumReact]

      implicit none

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      DECLARATION OF THE VARIABLES     !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

!      Allocatable arrays: declaration
       ! Integer: 1 bit
      integer*1, DIMENSION(:), allocatable :: StAlInst
       ! Integer: 2 bits
      integer*2, DIMENSION(:,:), allocatable :: OrgMetInst, OrgReactInst, in_nGen_nStru
      integer*4, DIMENSION(:,:), allocatable :: numRefInstLowHigh
       ! Character
      character(len=220), DIMENSION(:), allocatable :: mn, rn
       ! Logical
      logical, allocatable :: repeat

!      Static arrays: declaration
      ! Integer: 1 bit
      integer*1 :: InclRevStru(4000,4000)
      integer*1 :: EquiMetRevStru(28,4000), EquiReactRevStru(28,4000), SupEquiMetStru(8,4000)
      integer*1 :: SupEquiReactStru(8,4000), maxNumGen, maxNumMet, maxNumReact
      ! Integer: 2 bit
      integer*2 :: DataRevStru(12,4000)
      integer*2 :: nrefStruLow(15), nrefStruHigh(15), sup(5), n1, n2
      ! Integer: 4 bit
      integer*4 :: ResultsRevStruct(7,4000), MetRevStruct(1400,4000), ReactRevStruct(1400,4000)
      integer*4 :: MetABCRevStruct(1400,32000), React123RevStruct(1400,32000)
      integer*4 :: nrefInstLow(4000), nrefInstHigh(4000), EvaInstStru(4000)
      integer*4 :: supResultsRevStruct(4000), supRankStru(4000), const(10)
      integer*4 :: numInstReal, numStructures, ID, omp_get_thread_num
      integer*4 :: i, j, i1, i2, i3, i4, i5, nReact1, nReact2, nMet1, nMet2


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      INPUT FILES                      !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Reads the names of the reactions and of the metabolites and obtains the number of
!      metabolites and reactions
     !
!      Allocate the arrays mn(i1) and rn(i1), and the logical variable repeat
      allocate(mn(1400),rn(1400), repeat)
     !
     ! Metabolites
      !
      ! Logical variable. The loop is performed while the variable is "true"
      repeat = .true.
      ! Counts the number of metabolites
      n1 = 0
      open (1,file='List_Metabolites.txt')
      do while (repeat)
        ! Adds this new metabolite to the total
        n1 = n1 + 1
        ! Name of the metabolites
        read (1,1002) mn(n1)
        ! End the loop if the statement 'End' was found
        if( mn(n1) .eq. 'End' ) then
          ! Turn the logical variable into 'false' to end the loop
          repeat= .false.
          ! Subtract 1 to the total value of metabolites
          n1 = n1 - 1
        end if
      end do
      close (1)
      ! Deallocates the variable mn(i1)
      deallocate(mn)
     !
     ! Reactions
      !
      ! Logical variable. The cycle is performed while the variable is "true"
      repeat = .true.
      ! Counts the number of reactions (reversible and irreversible)
      n2 = 0
      open (1,file='List_Reactions.txt')
      do while (repeat)
        ! Adds this new reaction to the total
        n2 = n2 + 1
        ! Names of the unidirectional reactions
        read (1,1001) rn(n2)
        ! End the loop if the statement 'End' was found
        if( rn(n2) .eq. 'End' ) then
          ! Turn the logical variable into 'false' to end the loop
          repeat= .false.
          ! Subtract 1 to the total number of reactions
          n2 = n2 - 1
        end if
      end do
      close (1)
      ! Deallocates the variable rn(i1)
      deallocate(rn)


!      Reads the information regarding the possible structures till the number of generations
!      consideration (maxNumGen generations)
     !
    !  Read the file "in_nGen_nStru.txt"
      ! Allocate the variable in_nGen_nStru(i2,i1)
      allocate(in_nGen_nStru(2,15))
      !
      const(1) = 0
      ! Logical variable. The cycle is performed while the variable is "true"
      repeat = .true.
      open (1,file='in_nGen_nStru.txt')
      do while (repeat)
        const(1) = const(1) + 1
        read(1,*) (in_nGen_nStru(i1,const(1)), i1=1,2)
        if( const(1) .eq. (in_nGen_nStru(1,1)+1) ) then
          repeat= .false.
        end if
      end do
      close(1)
      const(1) = 0
      maxNumGen = in_nGen_nStru(1,1)
      numStructures = in_nGen_nStru(2,1)
      do i=2,(maxNumGen+1)
        nrefStruLow(i-1) = in_nGen_nStru(1,i)
        nrefStruHigh(i-1) = in_nGen_nStru(2,i)
      end do
      ! Deallocate the variables in_nGen_nStru(i2,i1) and repeat
      deallocate(in_nGen_nStru, repeat)
     !
    !  Read the file "in_nMet_nReact.txt", which stores  the maximum number of metabolites and of reactions considered
    !  for the structures within the assumed number of generations
      open (1,file='in_nMet_nReact.txt')
      do i1=1,2
        if( i1 .eq. 1 ) then
          read(1,*) maxNumMet
        else
          read(1,*) maxNumReact
        end if
      end do
      close(1)
     !
    !  Read the 'Structures_General.txt', which stores the general information regarding the structures (no restrictions)
      open (1,file='Structures_General.txt')
      do i1=1,numStructures
        read(1,*) (DataRevStru(i2,i1), i2=1,12)
      end do
      close(1)
     !
    !  Read the file 'EquiMetRevStru.txt', which stores the information regarding the equivalent metabolites of
    !  each structure
      open(1,file='EquiMetRevStru.txt')
      const(1) = 0
      do i1=1,(maxNumMet-1)
        const(1) = const(1) + (maxNumMet - i1)
      end do
      do i=1,numStructures
        read(1,*) (EquiMetRevStru(j,i), j=1,max(1,const(1)))
      end do
      close(1)
      !
    !  Read the file 'EquiReactRevStru.txt', which stores the information regarding the equivalent reactions of
    !  each structure
      open(1,file='EquiReactRevStru.txt')
      const(1) = 0
      do i1=1,(maxNumReact-1)
        const(1) = const(1) + (maxNumReact - i1)
      end do
      do i=1,numStructures
        read(1,*) (EquiReactRevStru(j,i), j=1,max(1,const(1)))
      end do
      close(1)
      !
    ! Read the file 'InclusionMatrix_Structures.txt', which stores the inclusion matrix. If the structure i1 can be
    ! found within structure i2, the corresponding cell is filled with "1"; if not, the cell is filled with "0"
      open(1,file='InclusionMatrix_Structures.txt')
      do i1=1,numStructures
        read(1,*) (InclRevStru(i2,i1), i2=1,numStructures)
      end do
      close(1)


!      Reads the information regarding the instances of the structures of the considered scope found
!      in the metabolic network
     !
    ! Reads the file 'numInstances.txt', which stores the number of instances found in the metabolic network
      open(1,file='numInstancesReal.txt')
      read(1,*) numInstReal
      close(1)
      !
    !
    ! Read the file 'OutRevStructReal.txt', which stores the naked frequency of the structures in the metabolic network
    ! and the number of different metabolites and reactions that appear in at least one instance of the corresponding
    ! structure
      open(1,file='OutRevStructReal.txt')
      do i1=1,numStructures
        read(1,*) ResultsRevStruct(1,i1)
      end do
      close(1)
      !
      ! Allocate the arrays OrgMetInst(i2,i1), OrgReactInst(i2,i1) and StAlInst(i1)
      allocate(OrgMetInst(maxNumMet,numInstReal),OrgReactInst(maxNumReact,numInstReal))
      allocate(StAlInst(numInstReal))
     !
    ! Reads the file 'OrgMetInst.txt', that stores the reference numbers of the metabolites of each instance
      open(1,file='OrgMetInst.txt')
      do i1=1,numInstReal
        read(1,*) (OrgMetInst(i2,i1), i2=1,maxNumMet)
      end do
      close(1)
     !
    ! Reads the file 'OrgReactInst.txt', that stores the reference numbers of the reactions of each instance
      open(1,file='OrgReactInst.txt')
      do i1=1,numInstReal
        read(1,*) (OrgReactInst(i2,i1), i2=1,maxNumReact)
      end do
      close(1)
     !
    ! Reads the file 'numRefInstLowHigh.txt', that stores lower and higher reference number of a instance of
    ! a given structure in organized matrices, OrgMetInst(i2,i1) and OrgReactInst(i2,i1), by structure to
    ! store the instances
      ! Allocate variable numRefInstLowHigh(i2,i1)
      allocate(numRefInstLowHigh(2,numStructures))
      open(1,file='numRefInstLowHigh.txt')
      do i1=1,numStructures
        read(1,*) (numRefInstLowHigh(i2,i1), i2=1,2)
      end do
      close(1)
      do i1=1,numStructures
        nrefInstLow(i1) = numRefInstLowHigh(1,i1)
        nrefInstHigh(i1) = numRefInstLowHigh(2,i1)
      end do
      ! Deallocate variable numRefInstLowHigh(i2,i1)
      deallocate(numRefInstLowHigh)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      EMBEDDED FREQUENCY               !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Obtains the embedded frequency for the structures
      do i1=1,numStructures
        !
        ! Adds the "naked" frequency of structure i1 to its "embedded" frequency
        ResultsRevStruct(2,i1) = ResultsRevStruct(1,i1)
        !
        ! Checks if the structure is not part of the last considered generation
        if( DataRevStru(1,i1) .ne. maxNumGen ) then
          ! Struture i1 is not part of the final generation
          !
          ! Checks the strutures i2 of higher generations which may have embedded
          ! the pattern of structure i1
          do i2=nrefStruLow( DataRevStru(1,i1) + 1 ),numStructures
            !
            ! The pattern of structure i1 can be found in structure i2?
            if( InclRevStru(i2,i1) .eq. 1 ) then
              ! Adds the "naked" frequency of structure i2 to the "embedded" frequency 5
              ! of structure i1
              ResultsRevStruct(2,i1) = ResultsRevStruct(2,i1) + ResultsRevStruct(1,i2)
            end if
            !
          end do
        end if
      end do


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      INSTANCES WITH NO OVERLAPS/      !
!      FREQUENCY OF THE METABOLITES     !
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Obtains the matrix SupEquiMetStru(i2,i1), which gives the information if a metabolite i2 is topologically
!      equivalent to another metabolite in a lower position
      ! Creates the matrices SupEquiMetStru(i2,i1) and SupEquiReactStru(i2,i1)
      do i1=1,numStructures
        do i2=1,maxNumMet
          SupEquiMetStru(i2,i1) = 0
        end do
        do i2=1,maxNumReact
          SupEquiReactStru(i2,i1) = 0
        end do
      end do
      !
      ! Checks for all the structures
      do i1=1,numStructures ! Number of structures
        !
        ! Metabolite A cannot be topologically equivalent to a metabolite in a lower position
        do 80 i2=2,DataRevStru(2,i1)  ! From the metabolite B to the metabolite DataRevStru(2,i1)
          !
          sup(1) = 0
          !
          ! Metabolite in position i2 can be topologically equivalent to a metabolite in a lower position?
          do i3=1,(i2 - 1)
            sup(2) = i2 - i3
            !
            ! Metabolite in position i2 is topologically equivalent to metabolite in position i3?
            if( EquiMetRevStru( (sup(1) + sup(2)) , i1 ) .eq. 1 ) then
              ! Metabolite i2 is topologically equivalent to a metabolite in a lower position
              SupEquiMetStru(i2,i1) = 1
              goto 80
            end if
            !
            sup(1) = sup(1) + ( DataRevStru(2,i1) - i3 )
            !
          end do
80      continue
        !
        ! Reaction r1 cannot be topologically equivalent to a reaction in a lower position
        ! Checks only for structures with 2 or more reactions
        if( DataRevStru(3,i1) .gt. 1 ) then
          do 85 i4=2,DataRevStru(3,i1)  ! From the metabolite B to the metabolite DataRevStru(3,i1)
            !
            sup(1) = 0
            !
            ! Reaction in position i4 can be topologically equivalent to a reaction in a lower position i5?
            do i5=1,(i4-1)
              sup(2) = i4 - i5
              !
              ! Reaction in position i4 is topologically equivalent to reaction in position i5?
              if( EquiReactRevStru( (sup(1) + sup(2)) , i1 ) .eq. 1 ) then
                ! Reaction i4 is topologically equivalent to a reaction in a lower position
                SupEquiReactStru(i4,i1) = 1
                goto 85
              end if
              !
              sup(1) = sup(1) + ( DataRevStru(3,i1) - i5 )
              !
            end do
85        continue
        end if
        !
      end do


!      Obtains the instances with no overlaps or with ovelaps of a maximum of 1 metabolite with
!      other instances of the same structure
      !
      write(*,*) 'Obtains the ranking of frequencies for the structures'
      write(*,*)
      ! Determines a ranking of frequency for the structures
      ! Obtains the variable supResultsRevStruct(i1)
      do i1=1,numStructures
        supResultsRevStruct(i1) = ResultsRevStruct(1,i1)
      end do
      ! Obtains the rank
      const(3) = 0
      do 90 i1=1,numStructures
        const(1) = 0
        const(2) = 0
        do i2=1,numStructures
          ! Structure i2 has a higher frequency than the frequency const(2) of the previous
          ! found structure const(1)?
          if( supResultsRevStruct(i2) .gt. const(2) )then
            const(2) = supResultsRevStruct(i2)
            const(1) = i2
          end if
          !
          ! Tested all the numStructures structures?
          if( i2 .eq. numStructures) then
            ! A structure was found with at least 1 instance?
            if( const(2) .ne. 0 ) then
              ! Stores the structure const(1) with the rank i1
              supRankStru(i1) = const(1)
              !
              ! Turns to "0" the frequency of structure const(1), to allow the determine the next
              ! structure with higher ranking
              supResultsRevStruct(const(1)) = 0
              !
              ! Search the structure with the next higher frequency
              goto 90
            else
              ! The structures that are still not ranked do not have instances
              !
              ! Stores the rank that should have been found in this iteration
              const(3) = i1
              !
              ! Leaves the loop
              goto 95
            end if
          end if
        end do
90    continue
      ! This part of the code is only available in the case all the structures have at least
      ! 1 instance. Goes to next part of the code
      goto 95
95    continue
      ! Is there at least 1 structure with no instances?
      if( const(3) .ne. 0 ) then
        ! Finds the rank of the structures with no instances
        do i1=1,numStructures
          ! Structure is has no instances?
          if( ResultsRevStruct(1,i1) .eq. 0 ) then
            supRankStru(const(3)) = i1
            const(3) = const(3) + 1
          end if
        end do
      end if
      !
      write(*,*) 'Starts finding the StAl instances'
      write(*,*)
      ! Creates the matrix StAlInst(i1), where: 0 => no overlaps; 1 => overlap 1 metabolite;
      ! 2 => shares reactions or two or more metabolites
      do i1=1,numInstReal
        StAlInst(i1) = 0
      end do
      ! Creates the matrix EvaInstStru(i1)
      do i1=1,numStructures
        EvaInstStru(i1) = 0
      end do
      !
      !
!      PARALLELIZATION USING OPENMP
      ! Parallelizes the loop
      ! Option: define the number of threads used?
  !    CALL OMP_SET_NUM_THREADS(2)
      !$omp parallel do default(shared) private(i,i1,i2,nMet1,nMet2,nReact1,nReact2,sup,const,ID) &
      !$omp schedule(dynamic)
      ! Checks the instances of the structure supRankStru(i)
      do i=1,numStructures
        !
        ! Structure supRankStru(i) has at least 1 instance?
        if( ResultsRevStruct(1,supRankStru(i)) .ne. 0 ) then
          ! Structure supRankStru(i) has at least 1 instance
          !
          ! Structure supRankStru(i) has more than 1 instance?
          if( ResultsRevStruct(1,supRankStru(i)) .gt. 1 ) then
            !
            ! Checks all the possible pairs of instances of structure supRankStru(i)
            do i1=nrefInstLow(supRankStru(i)),(nrefInstHigh(supRankStru(i))-1) ! From nrefInstLow(supRankStru(i))
              !                                                                ! to nrefInstHigh(supRankStru(i))-1
              ! Determines the frequency of the specific metabolites and reactions in each one of the structures
              ! Metabolites
              ! Call subroutine freqNodesInst, to add to the frequency in structure supRankStru(i) the metabolites
              ! of instance i1
              sup(1) = 6  ! Needed to obtain the number of different specific metabolites in structure supRankStru(i)
              call freqNodesInst(supRankStru(i), DataRevStru(2,supRankStru(i)), i1, OrgMetInst, &
              maxNumMet, numInstReal, SupEquiMetStru, EquiMetRevStru, sup(1), MetRevStruct, &
              MetABCRevStruct, ResultsRevStruct)
              !
              ! Reactions
              ! Call subroutine freqNodesInst, to add to the frequency in structure supRankStru(i) the reactions
              ! of instance i1
              sup(1) = 7  ! Needed to obtain the number of different specific reactions in structure supRankStru(i)
              call freqNodesInst(supRankStru(i), DataRevStru(3,supRankStru(i)), i1, OrgReactInst, &
              maxNumReact, numInstReal, SupEquiReactStru, EquiReactRevStru, sup(1), ReactRevStruct, &
              React123RevStruct, ResultsRevStruct)
              !
              !
              ! Checks if all the instances of structure supRankStru(i) were all permanentely defined as having
              ! or not overlaps
              if( EvaInstStru(supRankStru(i)) .ne. ResultsRevStruct(1,supRankStru(i)) ) then
                ! Not all the instances of structure supRankStru(i) have been permanentely defined as having or
                ! not overlaps
                !
                ! Compares instance i1 to the instances in positions equal or higher than (i1+1)
                do 100 i2=(i1+1),max((i1+1),nrefInstHigh(supRankStru(i)))  ! (i1+1) to maximum reference number
                  !                                        ! of a instance of the same structure of instance i1
                  ! Checks if both instances i1 and i2 have overlaps
                  if( (StAlInst(i1) .eq. 2) .and. (StAlInst(i2) .eq. 2) ) goto 100
                  !
                  ! Checks if the instances i1 and i2 have at least one reaction in common
                  do nReact1=1,DataRevStru(3,supRankStru(i))  ! Number of reactions of the instances
                    do nReact2=1,DataRevStru(3,supRankStru(i))  ! Number of reactions of the instances
                      !
                      ! Checks if reaction OrgReactInst(nReact1,i1) is equal to reaction OrgReactInst(nReact2,i2)
                      if( OrgReactInst(nReact1,i1) .eq. OrgReactInst(nReact2,i2) ) then
                        ! It is the same reaction
                        !
                        ! Changes the information in matrix StAlInst(i1) to indicate the instances have overlaps
                        ! Instance i1
                        if( StAlInst(i1) .ne. 2 ) then
                          StAlInst(i1) = 2
                          EvaInstStru(supRankStru(i)) = EvaInstStru(supRankStru(i)) + 1
                        end if
                        ! Instance i2
                        if( StAlInst(i2) .ne. 2 ) then
                          StAlInst(i2) = 2
                          EvaInstStru(supRankStru(i)) = EvaInstStru(supRankStru(i)) + 1
                        end if
                        goto 100
                      end if
                    end do
                  end do
                  !
                  const(1) = 0
                  ! Checks the number of common metabolites of instance i1 and i2
                  do nMet1=1,DataRevStru(2,supRankStru(i))  ! Number of metabolites of the instances
                    do nMet2=1,DataRevStru(2,supRankStru(i))  ! Number of metabolites of the instances
                      !
                      ! Metabolites OrgMetInst(nMet1,i1) and OrgMetInst(nMet2,i2) are the same?
                      if( OrgMetInst(nMet1,i1) .eq. OrgMetInst(nMet2,i2) ) const(1) = const(1) + 1
                      !
                      ! Checks if the instances i1 and i2 share more than 1 metabolite
                      if( const(1) .gt. 1 ) then
                        ! Instances i1 and i2 have overlaps
                        if( StAlInst(i1) .ne. 2 ) then
                          StAlInst(i1) = 2
                          EvaInstStru(supRankStru(i)) = EvaInstStru(supRankStru(i)) + 1
                        end if
                        if( StAlInst(i2) .ne. 2 ) then
                          StAlInst(i2) = 2
                          EvaInstStru(supRankStru(i)) = EvaInstStru(supRankStru(i)) + 1
                        end if
                        goto 100
                      end if
                      !
                      ! Checks if all the metabolites were compared
                      if( (nMet1 .eq. DataRevStru(2,supRankStru(i))) .and. &
                      (nMet2 .eq. DataRevStru(2,supRankStru(i))) ) then
                        if( const(1) .ne. 0 ) then
                          ! Check if instances i1 and i2 are still considered with no overlaps
                          if( StAlInst(i1) .eq. 0 ) StAlInst(i1) = 1
                          if( StAlInst(i2) .eq. 0 ) StAlInst(i2) = 1
                        end if
                      end if
                      !
                    end do
                  end do
100             continue
                !
                ! Checks if the instance i1 has overlaps
                if( StAlInst(i1) .ne. 2 ) then
                  !
                  ! Adds 1 to the total number of instances with no overlaps or with a maximum of a overlap of 1 metabolite
                  ResultsRevStruct(5,supRankStru(i)) = ResultsRevStruct(5,supRankStru(i)) + 1
                  EvaInstStru(supRankStru(i)) = EvaInstStru(supRankStru(i)) + 1
                  !
                  if( StAlInst(i1) .eq. 0 ) then
                    ! Type i: no overlaps
                    !
                    ! Adds 1 to the total number of instances with no overlaps of structure supRankStru(i)
                    ResultsRevStruct(3,supRankStru(i)) = ResultsRevStruct(3,supRankStru(i)) + 1
                    !
                  else
                    ! Type ii: overlap of 1 metabolite
                    !
                    ! Adds 1 to the total number of instances with a maximum of a overlap of 1 metabolite of
                    ! structure supRankStru(i)
                    ResultsRevStruct(4,supRankStru(i)) = ResultsRevStruct(4,supRankStru(i)) + 1
                    !
                  end if
                  !
                end if
              end if
            end do
          end if
          !
          ! Counts the number of metabolites and of reactions of the last instance nrefInstHigh(supRankStru(i))
          ! of structure supRankStru(i); determines if instance nrefInstHigh(supRankStru(i)) has overlapps
          !
          ! Metabolites
          ! Call subroutine freqNodesInst, to add to the frequency in structure supRankStru(i) the metabolites
          ! of instance nrefInstHigh(supRankStru(i))
          sup(1) = 6  ! Needed to obtain the number of different specific metabolites in structure supRankStru(i)
          call freqNodesInst(supRankStru(i), DataRevStru(2,supRankStru(i)), &
          nrefInstHigh(supRankStru(i)), OrgMetInst, maxNumMet, numInstReal, SupEquiMetStru, &
          EquiMetRevStru, sup(1), MetRevStruct, MetABCRevStruct, ResultsRevStruct)
          !
          ! Reactions
          ! Call subroutine freqNodesInst, to add to the frequency in structure supRankStru(i) the reactions
          ! of instance nrefInstHigh(supRankStru(i))
          sup(1) = 7  ! Needed to obtain the number of different specific reactions in structure supRankStru(i)
          call freqNodesInst(supRankStru(i), DataRevStru(3,supRankStru(i)), &
          nrefInstHigh(supRankStru(i)), OrgReactInst, maxNumReact, numInstReal, SupEquiReactStru, &
          EquiReactRevStru, sup(1), ReactRevStruct, React123RevStruct, ResultsRevStruct)
          !
          ! Checks if the instance nrefInstHigh(supRankStru(i)) has overlaps
          if( StAlInst(nrefInstHigh(supRankStru(i))) .ne. 2 ) then
            !
            ! Adds 1 to the total number of instances with no overlaps or with a maximum of a overlap
            ! of 1 metabolite
            ResultsRevStruct(5,supRankStru(i)) = ResultsRevStruct(5,supRankStru(i)) + 1
            EvaInstStru(supRankStru(i)) = EvaInstStru(supRankStru(i)) + 1
            !
            if( StAlInst(nrefInstHigh(supRankStru(i))) .eq. 0 ) then
              ! Type i: no overlaps
              !
              ! Adds 1 to the total number of instances with no overlaps of structure supRankStru(i)
              ResultsRevStruct(3,supRankStru(i)) = ResultsRevStruct(3,supRankStru(i)) + 1
              !
            else
              ! Type ii: overlap of 1 metabolite
              !
              ! Adds 1 to the total number of instances with a maximum of a overlap of 1 metabolite of
              ! structure supRankStru(i)
              ResultsRevStruct(4,supRankStru(i)) = ResultsRevStruct(4,supRankStru(i)) + 1
              !
            end if
            !
          end if
        end if
        ! Stores the ID number of the the thread used
        ID = omp_get_thread_num()
        write(*,1004) supRankStru(i), i, ResultsRevStruct(1,supRankStru(i)), &
        ResultsRevStruct(3,supRankStru(i)), ResultsRevStruct(4,supRankStru(i)), ID
      end do
!     END PARALLELIZATION
      !
      const(2) = 0
      const(3) = 0
      const(4) = 0
      const(5) = 0
      do i=1,numStructures
        const(2) = const(2) + ResultsRevStruct(3,i)
        const(3) = const(3) + ResultsRevStruct(4,i)
        const(4) = const(4) + ResultsRevStruct(5,i)
      end do
      const(5) = numInstReal - const(4)
      write(*,*)
      write(*,*) 'Instances with overlaps determined:'
      write(*,1005) numInstReal, const(5), const(2), const(3)



      open(8,file='StAlInst.txt')
      do i1=1,numInstReal
        write(8,1003) StAlInst(i1)
      end do
      close(8)
      ! Deallocates the variables StAlInst(i1), OrgMetInst(i2,i1) and OrgReactInst(i2,i1)
      deallocate(StAlInst)
      deallocate(OrgMetInst)
      deallocate(OrgReactInst)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Stores the general information obtained for the structures, such as: "naked"
!      frequency, "embedded" frequency, number of stand alone-alone instances (type i,
!      0 metabolites + 0 reactions; type ii, 1 metabolite + 0 reactions; and
!      type i + ii), and number of different metabolites and reactions. File:
!      'ResultsRevStruct.txt.txt'
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='ResultsRevStruct.txt')
      do i1=1,numStructures
        write(8,1006) (ResultsRevStruct(i2,i1), i2=1,7)
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Stores the frequency of the metabolites in the instances of the structures of
!      the considered scope
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='MetRevStruct.txt')
      do i1=1,numStructures
        write(8,1006) (MetRevStruct(i2,i1), i2=1,n1)
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Stores the frequency of the reactions in the instances of the structures of
!      the considered scope
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='ReactRevStruct.txt')
      do i1=1,numStructures
        write(8,1006) (ReactRevStruct(i2,i1), i2=1,n2)
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Stores the frequency of the metabolites in all the possible positions of
!      the instances (A, B, C, ...) of the structures of the considered scope
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='MetABCRevStruct.txt')
      do i1=1,(numStructures*maxNumMet)
        write(8,1006) (MetABCRevStruct(i2,i1), i2=1,n1)
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Stores the frequency of the reactions in all the possible positions of
!      the instances (r1, r2, r3, ...) of the structures of the considered scope
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      open(8,file='React123RevStruct.txt')
      do i1=1,(numStructures*maxNumReact)
        write(8,1006) (React123RevStruct(i2,i1), i2=1,n2)
      end do
      close(8)


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!      Formats
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
1001  format (a220)
1002  format (a100)
1003  format (6000(i6, ','))
1004  format('Nstru=', i5, ';  Nrank=', i5, ';  Ninst=', i10, ';  NsaI=', i4, &
             ';  NsaII=', i4, ';  Nthread=', i3)
1005  format (' Ninst=', i10, ';   N.ol=', i10, ';   NsaI=', i5, ';   NsaII=', i5)
1006  format (6000(i10, ','))


      contains


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!     Subroutine to obtain the frequency of the specific nodes (metabolites and
!!     reactions) in the instances of each structure for two situations:
!!     i) the position in which the specific node is found in the instance is not
!!     taken into account; ii) the frequency in each of the possible positions
!!     is determined (topologically equivalent node positions are considered
!!     the same position)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      subroutine freqNodesInst(numRefStr_, numNodes_, numRefInst_, OrgNodeInst_, &
      maxNumNodes_, numInstReal_, SupEquiNodeStru_, EquiNodeRevStru_, inDifNodes_, &
      NodeRevStruct_, Node123RevStruct_, ResultsRevStruct_)

! CONSTANTS
!      numNodes_ : number of nodes, metabolites or reactions, of structure numRefStr_ (given by
!                  the variables DataRevStru(2,i1) and DataRevStru(3,i1), repectively, of the
!                  main program)
!      numRefStr_ : reference number of the structure of instance numRefInst_
!      numRefInst_ : reference number of the considered instance
!      maxNumNodes_ : maximum number of nodes (metabolites or reactions) considered for the structures within the
!                     considered number of generations maxNumGen
!      numInstReal_ :  number of instances for the considered scope
!      inDifNodes_ : input integer needed to determine the number of different specific nodes (metabolites or
!                    reactions) present in instances of structure numRefStr_. For metabolites, inDifNodes_ = 6;
!                    for reactions, inDifNodes_ = 7
!
!
! VECTORS
!      sup_(i1) : support vector of size integer*2
!
!
! MATRICES
!      OrgNodeInst_(i2,i1) : reference numbers of the specific nodes (metabolites or reactions) of the instance i1. The instances
!                            are organized according to their structure: instances of structures of lower reference numbers are
!                            stored first
!      SupEquiNodeStru_(i2,i1) : support matrix that gives the information if the node i2 (metabolite or reaction) is topologically
!                                equivalent to another node in a lower position of structure i1
!      EquiNodeRevStru_(i2,i1) : information regarding the equivalent nodes (metabolites or reactions) of a given structure i1.
!                                If two nodes are equivalent, the corresponding cell is filled with "1"; if not, is filled
!                                with "0". The first node is compared with the ones in higher positions, the second node is
!                                compared to nodes who are also in higher positions, ... till we compare the last two nodes.
!                                For the node in a general position X (lower than the last position available), it will
!                                be compared to nodes in increasing higher positions in the cell. Considering the case
!                                where the node in the position X is compared to the node in the position Y, the
!                                information regarding the topological equivalence of the two nodes will be stored in the cell:
!                                i2 = (Somatrio[i=1,(X-1)] n Node - i ) + ( Y - X )
!      ResultsRevStruct_(i2,i1) : output data obtained for the structure i1 with the developed code. (i2=1) "naked" frequency (number of
!                                 instances); (i2=2) "embedded" frequency; (i2=3) n of stand-alone instances of type i - 0 metabolites
!                                 and 0 reactions shared with other instances of the same structure; (i2=4) n of stand-alone instances
!                                 of type ii - 1 metabolite and 0 reactions shared with other instances of the same structure; (i2=5) total
!                                 n stand-alone instances - type i + ii; (i2=6) n of different metabolites; (i2=7) n of different reactions
!      NodeRevStruct_(i2,i1) : frequency of the specific node (metabolite or reaction) i2 in the instances of structure i1
!      Node123RevStruct_(i2,i1) : frequency of the specific node (metabolite or reaction) i2 in the instances of a given structure
!                                 at all the possible positions. For a given structure X, the frequency of the specific node
!                                 i2 in the position Y is: i1 = [(X - 1) x maxNumNode_] + Y, where Y => [1,maxNumNode_]

      implicit none

      ! IN variables
       ! Static variables
        ! Integer: 1 bit
      integer*1, intent(in) :: SupEquiNodeStru_(:,:), EquiNodeRevStru_(:,:)
      integer*1, intent(in) :: maxNumNodes_
        ! Integer: 2 bits
      integer*2, intent(in) :: numNodes_, inDifNodes_
        ! Integer: 4 bits
      integer*4, intent(in) :: numRefStr_, numRefInst_, numInstReal_
       ! Allocatable arrays: declaration
        ! Integer: 2 bits
      integer*2, DIMENSION(maxNumNodes_,numInstReal_), intent(in) :: OrgNodeInst_

      ! IN/OUT variables
       ! Static variables
        ! Integer: 4 bits
      integer*4, intent(inout) :: NodeRevStruct_(:,:), Node123RevStruct_(:,:)
      integer*4, intent(inout) :: ResultsRevStruct_(:,:)

      ! Internal support variables
       ! Static variables
        ! Integer: 2 bits
      integer*2 :: sup_(2)

      ! General variables
      integer*4 :: iF1_, iF2_

!      Adds to the frequency of the specific nodes (metabolites or reactions) of instance numRefInst_
!      in structure numRefStr_
      do 500 iF1_=1,numNodes_  ! Number of nodes (metabolites or reactions) of instance numRefInst_

        ! Checks if this metabolite was found before in instances of structure numRefStr_
        if( NodeRevStruct_( OrgNodeInst_(iF1_,numRefInst_) , numRefStr_ ) .eq. 0 ) then
          ! Adds 1 to the total number of different metabolites of structure numRefStr_
          ResultsRevStruct_(inDifNodes_,numRefStr_) = ResultsRevStruct_(inDifNodes_,numRefStr_) + 1
        end if

        ! Adds 1 to the total frequency of the specific node (metabolite or reactions) OrgNodeInst_(iF1_,numRefInst_)
        ! in instances of structure numRefStr_
        NodeRevStruct_( OrgNodeInst_(iF1_,numRefInst_) , numRefStr_ ) = &
        NodeRevStruct_( OrgNodeInst_(iF1_,numRefInst_) , numRefStr_ ) + 1
        !
        ! Counts the specific node (metabolite or reaction) OrgNodeInst_(iF1_,numRefInst_) taking into account
        ! its position
        ! The position iF1_ is topologically equivalent to a lower position?
        if( SupEquiNodeStru_(iF1_,numRefStr_) .eq. 1 ) then
          ! Position iF1_ is topologically equivalent to a lower position
          !
          sup_(1) = 0
          !
          ! Find the lower position to which the position iF1_ is topologically equivalent
          do iF2_=1,(iF1_ - 1)  ! From 1 to the position (iF1_ - 1)
            sup_(2) = iF1_ - iF2_
            !
            ! Position iF1_ is topologically equivalent to the lower position iF2_?
            if( EquiNodeRevStru_( (sup_(1) + sup_(2)) , numRefStr_ ) .eq. 1 ) then
              ! Position iF1_ is equivalent to position iF2_
              !
              ! Adds 1 to the frequency of the specific node (metabolite or reaction) OrgNodeInst_(iF1_,numRefInst_)
              ! in the position iF2_
              Node123RevStruct_( OrgNodeInst_(iF1_,numRefInst_) , ( ((numRefStr_ - 1)*maxNumNodes_) + iF2_ ) ) = &
              Node123RevStruct_( OrgNodeInst_(iF1_,numRefInst_) , ( ((numRefStr_ - 1)*maxNumNodes_) + iF2_ ) ) + 1
              goto 500
              !
            end if
            !
            sup_(1) = sup_(1) + ( numNodes_ - iF2_ )
            !
          end do
          !
        else
          ! Position iF1_ is not topologically equivalent to a lower position
          !
          ! Adds 1 to the frequency of the specific node (metabolite or reaction OrgNodeInst_(iF1_,numRefInst_)
          ! in the position iF1_
          Node123RevStruct_( OrgNodeInst_(iF1_,numRefInst_) , ( ((numRefStr_ - 1)*maxNumNodes_) + iF1_ ) ) = &
          Node123RevStruct_( OrgNodeInst_(iF1_,numRefInst_) , ( ((numRefStr_ - 1)*maxNumNodes_) + iF1_ ) ) + 1
          !
        end if
        !
500   continue

      end subroutine freqNodesInst


      end program StAlInstCount
